'use strict';

let https = require('https');
let url = require('url');
let querystring = require('querystring');
let fs = require('fs');
let process = require('process');

const BASE_URL = "https://api.fax2.com.au/v1/";

class Fax2 {
	static now() { return Date.now() / 1000; }
	
	static date_as_iso8601(d) {
		return d.toISOString().replace(/\.\d\d\d/,'');
	}
	
	static date_from_iso8601(s) { return new Date(s); }
	
	static date_add_seconds(d,i) {
		return new Date(d.getTime() + i * 1000);
	}
	
	constructor() {
		this.token_expires_at = 0;
	}
	
	https_req(_url,args) {
		let url_args = url.parse(_url);
		let https_args = {
			'hostname': url_args.hostname,
			'port': url_args.port || 443,
			'path': url_args.path
		};
		https_args = Object.assign(https_args,args);
		if (https_args.query !== undefined) {
			// This will break if there are query string parameters
			// in the URL already, so we don't do that.
			let query_string = querystring.stringify(https_args.query)
			delete https_args.query
			if(https_args.method === 'POST') {
				https_args.data = query_string;
			} else {
				https_args.path.concat('?' + query_string);
			}
		}
		
		return new Promise((resolve,reject) => {
			let resp_body = '';
			let req = https.request(https_args,(res) => {
				res.setEncoding('binary');
				res.on('data', (data) => { resp_body = resp_body + data; });
				res.on('end', () => {
					if(Math.floor(res.statusCode/100) == 2) {
						resolve(resp_body);
					} else {
						reject(JSON.parse(resp_body));
					}
				});
			});
			
			if(https_args.data !== undefined) {
				req.write(https_args.data,undefined,() => req.end());
			} else req.end();
		});
	}
	
	get_auth_token() {
		return new Promise((resolve,reject) => {
			this.https_req(
				BASE_URL + 'oauth2/token',
				{
					auth: `${this.username}:${this.password}`,
					method: 'POST',
					query: {
						grant_type: 'client_credentials'
					}
				}
			).then((data) => {
				data = JSON.parse(data);
				this.token_expires_at = Fax2.date_add_seconds(new Date(),data.expires_in);
				resolve(this.token = data.access_token);
			}).catch((e) => { reject(e); });
		});
	}
	
	get_auth_token_maybe() {
		if (this.token_expires_at < Fax2.now() + 60)
			return this.get_auth_token();
		else
			return new Promise((resolve,reject) => resolve());
	}

	auth_https_req(_url,args) {
		if(args === undefined) args = {};
		// Assume no heaeders in incoming args
		// This is the only header we use anyway.
		args = Object.assign({},args);
		args.headers = Object.assign({},args.headers);
		Object.assign(args.headers,{ 'Authorization': `bearer ${this.token}` });
		
		return this.https_req(_url,args);
	}
	
	url_req(_url,args) {
		return this.get_auth_token_maybe().
			then(() => this.auth_https_req(_url,args));
	}
	
	req(endpoint,args) {
		return this.url_req(BASE_URL + endpoint,args);
	}
	
	json_url_req(url,args) {
		return new Promise((resolve,reject) => {
			this.url_req(url,args).then((resp) => {
				resolve(JSON.parse(resp));
			}).catch((e) => reject(e));
		});
	}
	
	json_req(endpoint,args) {
		return this.json_url_req(BASE_URL + endpoint,args);
	}
	
	received_faxes_page(accum,url) {
		return new Promise((resolve,reject) => {
			this.json_url_req(url).then((data) => {
				let list = data['data'];
				for(let fax of list)
					fax.received_at = Fax2.date_from_iso8601(fax.received_at);
				accum = accum.concat(list);
				if(data["next_page_url"]) {
					this.received_faxes_page(accum,data["next_page_url"]).
						then((ret) => resolve(ret));
				} else {
					resolve(accum);
				}
			}).catch((e) => { reject(e); });
		});
	}

	received_faxes(args) {
		if(args === undefined) args = {};
		let req_args = {};
		if(args.from_time)
			req_args.from_time = Fax2.date_as_iso8601(args.from_time);
		
		let url = BASE_URL + 'received_faxes?' +
			querystring.stringify(req_args);
		return this.received_faxes_page([],url);
	}
	
	upload_document(document_data,mime_type) {
		return this.json_req('upload_document',
			{
				method: 'POST',
				headers: {"Content-Type":mime_type},
				data: document_data
			}
		).then((data) => {
			return [data.document_id,data.pages];
		});
	}
	
	send_fax(dest_number,documents) {
		return this.json_req(
			'send_fax', {
				method: 'POST',
				query: {
					'dest_number': dest_number,
					'documents[]': documents
				}
			}
		).then((data) => {
			return [data.id,data.pages];
		});
	}
	
	get_received_fax_pdf_data(id) {
		return this.req(`received_faxes/${id}/content.pdf`);
	}
	
	write_received_fax_pdf(id,filename) {
		return this.get_received_fax_pdf_data(id).then((data) => {
			let tempname = `${filename}.${process.pid}`
			fs.writeFileSync(tempname,data,'binary');
			fs.renameSync(tempname,filename);
		});
	}
}

module.exports = Fax2;
